<?php
/**
 * @copyright Copyright (C) 2010-2022, the Friendica project
 *
 * @license GNU AGPL version 3 or any later version
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */

namespace Friendica\Model;

use DOMDocument;
use DOMXPath;
use Exception;
use Friendica\Core\Logger;
use Friendica\Core\Protocol;
use Friendica\Core\System;
use Friendica\Core\Worker;
use Friendica\Database\Database;
use Friendica\Database\DBA;
use Friendica\DI;
use Friendica\Module\Register;
use Friendica\Network\HTTPClient\Client\HttpClientAccept;
use Friendica\Network\HTTPClient\Client\HttpClientOptions;
use Friendica\Network\HTTPClient\Capability\ICanHandleHttpResponses;
use Friendica\Network\Probe;
use Friendica\Protocol\ActivityPub;
use Friendica\Protocol\Relay;
use Friendica\Util\DateTimeFormat;
use Friendica\Util\JsonLD;
use Friendica\Util\Network;
use Friendica\Util\Strings;
use Friendica\Util\XML;
use Friendica\Network\HTTPException;
use GuzzleHttp\Psr7\Uri;

/**
 * This class handles GServer related functions
 */
class GServer
{
	// Directory types
	const DT_NONE = 0;
	const DT_POCO = 1;
	const DT_MASTODON = 2;

	// Methods to detect server types

	// Non endpoint specific methods
	const DETECT_MANUAL = 0;
	const DETECT_HEADER = 1;
	const DETECT_BODY = 2;
	const DETECT_HOST_META = 3;
	const DETECT_CONTACTS = 4;
	const DETECT_AP_ACTOR = 5;
	const DETECT_AP_COLLECTION = 6;

	const DETECT_UNSPECIFIC = [self::DETECT_MANUAL, self::DETECT_HEADER, self::DETECT_BODY, self::DETECT_HOST_META, self::DETECT_CONTACTS, self::DETECT_AP_ACTOR];

	// Implementation specific endpoints
	// @todo Possibly add Lemmy detection via the endpoint /api/v3/site
	const DETECT_FRIENDIKA = 10;
	const DETECT_FRIENDICA = 11;
	const DETECT_STATUSNET = 12;
	const DETECT_GNUSOCIAL = 13;
	const DETECT_CONFIG_JSON = 14; // Statusnet, GNU Social, Older Hubzilla/Redmatrix
	const DETECT_SITEINFO_JSON = 15; // Newer Hubzilla
	const DETECT_MASTODON_API = 16;
	const DETECT_STATUS_PHP = 17; // Nextcloud
	const DETECT_V1_CONFIG = 18;
	const DETECT_PUMPIO = 19; // Deprecated
	const DETECT_SYSTEM_ACTOR = 20; // Mistpark, Osada, Roadhouse, Zap

	// Standardized endpoints
	const DETECT_STATISTICS_JSON = 100;
	const DETECT_NODEINFO_1 = 101;
	const DETECT_NODEINFO_2 = 102;
	const DETECT_NODEINFO_210 = 103;

	/**
	 * Check for the existance of a server and adds it in the background if not existant
	 *
	 * @param string $url
	 * @param boolean $only_nodeinfo
	 *
	 * @return void
	 */
	public static function add(string $url, bool $only_nodeinfo = false)
	{
		if (self::getID($url, false)) {
			return;
		}

		Worker::add(Worker::PRIORITY_LOW, 'UpdateGServer', $url, $only_nodeinfo);
	}

	/**
	 * Get the ID for the given server URL
	 *
	 * @param string $url
	 * @param boolean $no_check Don't check if the server hadn't been found
	 *
	 * @return int|null gserver id or NULL on empty URL or failed check
	 */
	public static function getID(string $url, bool $no_check = false): ?int
	{
		$url = self::cleanURL($url);

		if (empty($url)) {
			return null;
		}

		$gserver = DBA::selectFirst('gserver', ['id'], ['nurl' => Strings::normaliseLink($url)]);
		if (DBA::isResult($gserver)) {
			Logger::debug('Got ID for URL', ['id' => $gserver['id'], 'url' => $url, 'callstack' => System::callstack(20)]);
			return $gserver['id'];
		}

		if ($no_check || !self::check($url)) {
			return null;
		}

		return self::getID($url, true);
	}

	/**
	 * Retrieves all the servers which base domain are matching the provided domain pattern
	 *
	 * The pattern is a simple fnmatch() pattern with ? for single wildcard and * for multiple wildcard
	 *
	 * @param string $pattern
	 *
	 * @return array
	 *
	 * @throws Exception
	 */
	public static function listByDomainPattern(string $pattern): array
	{
		$likePattern = 'http://' . strtr($pattern, ['_' => '\_', '%' => '\%', '?' => '_', '*' => '%']);

		// The SUBSTRING_INDEX returns everything before the eventual third /, which effectively trims an
		// eventual server path and keep only the server domain which we're matching against the pattern.
		$sql = "SELECT `gserver`.*, COUNT(*) AS `contacts`
			FROM `gserver`
			LEFT JOIN `contact` ON `gserver`.`id` = `contact`.`gsid`
			WHERE SUBSTRING_INDEX(`gserver`.`nurl`, '/', 3) LIKE ?
			AND NOT `gserver`.`failed`
			GROUP BY `gserver`.`id`";

		$stmt = DI::dba()->p($sql, $likePattern);

		return DI::dba()->toArray($stmt);
	}

	/**
	 * Checks if the given server is reachable
	 *
	 * @param string  $profile URL of the given profile
	 * @param string  $server  URL of the given server (If empty, taken from profile)
	 * @param string  $network Network value that is used, when detection failed
	 * @param boolean $force   Force an update.
	 *
	 * @return boolean 'true' if server seems vital
	 */
	public static function reachable(string $profile, string $server = '', string $network = '', bool $force = false): bool
	{
		if ($server == '') {
			$contact = Contact::getByURL($profile, null, ['baseurl']);
			if (!empty($contact['baseurl'])) {
				$server = $contact['baseurl'];
			}
		}

		if ($server == '') {
			return true;
		}

		return self::check($server, $network, $force);
	}

	/**
	 * Calculate the next update day
	 *
	 * @param bool $success
	 * @param string $created
	 * @param string $last_contact
	 * @param bool $undetected
	 *
	 * @return string
	 *
	 * @throws Exception
	 */
	public static function getNextUpdateDate(bool $success, string $created = '', string $last_contact = '', bool $undetected = false): string
	{
		// On successful contact process check again next week when it is a detected system.
		// When we haven't detected the system, it could be a static website or a really old system.
		if ($success) {
			return DateTimeFormat::utc($undetected ? 'now +1 month' : 'now +7 day');
		}

		$now = strtotime(DateTimeFormat::utcNow());

		if ($created > $last_contact) {
			$contact_time = strtotime($created);
		} else {
			$contact_time = strtotime($last_contact);
		}

		// If the last contact was less than 6 hours before then try again in 6 hours
		if (($now - $contact_time) < (60 * 60 * 6)) {
			return DateTimeFormat::utc('now +6 hour');
		}

		// If the last contact was less than 12 hours before then try again in 12 hours
		if (($now - $contact_time) < (60 * 60 * 12)) {
			return DateTimeFormat::utc('now +12 hour');
		}

		// If the last contact was less than 24 hours before then try tomorrow again
		if (($now - $contact_time) < (60 * 60 * 24)) {
			return DateTimeFormat::utc('now +1 day');
		}

		// If the last contact was less than a week before then try again in a week
		if (($now - $contact_time) < (60 * 60 * 24 * 7)) {
			return DateTimeFormat::utc('now +1 week');
		}

		// If the last contact was less than two weeks before then try again in two week
		if (($now - $contact_time) < (60 * 60 * 24 * 14)) {
			return DateTimeFormat::utc('now +2 week');
		}

		// If the last contact was less than a month before then try again in a month
		if (($now - $contact_time) < (60 * 60 * 24 * 30)) {
			return DateTimeFormat::utc('now +1 month');
		}

		// The system hadn't been successul contacted for more than a month, so try again in three months
		return DateTimeFormat::utc('now +3 month');
	}

	/**
	 * Checks the state of the given server.
	 *
	 * @param string  $server_url    URL of the given server
	 * @param string  $network       Network value that is used, when detection failed
	 * @param boolean $force         Force an update.
	 * @param boolean $only_nodeinfo Only use nodeinfo for server detection
	 *
	 * @return boolean 'true' if server seems vital
	 */
	public static function check(string $server_url, string $network = '', bool $force = false, bool $only_nodeinfo = false): bool
	{
		$server_url = self::cleanURL($server_url);
		if ($server_url == '') {
			return false;
		}

		$gserver = DBA::selectFirst('gserver', [], ['nurl' => Strings::normaliseLink($server_url)]);
		if (DBA::isResult($gserver)) {
			if ($gserver['created'] <= DBA::NULL_DATETIME) {
				$fields = ['created' => DateTimeFormat::utcNow()];
				$condition = ['nurl' => Strings::normaliseLink($server_url)];
				self::update($fields, $condition);
			}

			if (!$force && (strtotime($gserver['next_contact']) > time())) {
				Logger::info('No update needed', ['server' => $server_url]);
				return (!$gserver['failed']);
			}
			Logger::info('Server is outdated. Start discovery.', ['Server' => $server_url, 'Force' => $force]);
		} else {
			Logger::info('Server is unknown. Start discovery.', ['Server' => $server_url]);
		}

		return self::detect($server_url, $network, $only_nodeinfo);
	}

	/**
	 * Set failed server status
	 *
	 * @param string $url
	 */
	public static function setFailure(string $url)
	{
		$gserver = DBA::selectFirst('gserver', [], ['nurl' => Strings::normaliseLink($url)]);
		if (DBA::isResult($gserver)) {
			$next_update = self::getNextUpdateDate(false, $gserver['created'], $gserver['last_contact']);
			self::update(['url' => $url, 'failed' => true, 'last_failure' => DateTimeFormat::utcNow(),
			'next_contact' => $next_update, 'network' => Protocol::PHANTOM, 'detection-method' => null],
			['nurl' => Strings::normaliseLink($url)]);
			Logger::info('Set failed status for existing server', ['url' => $url]);
			return;
		}
		self::insert(['url' => $url, 'nurl' => Strings::normaliseLink($url),
			'network' => Protocol::PHANTOM, 'created' => DateTimeFormat::utcNow(),
			'failed' => true, 'last_failure' => DateTimeFormat::utcNow()]);
		Logger::info('Set failed status for new server', ['url' => $url]);
	}

	/**
	 * Remove unwanted content from the given URL
	 *
	 * @param string $url
	 *
	 * @return string cleaned URL
	 */
	public static function cleanURL(string $url): string
	{
		$url = trim($url, '/');
		$url = str_replace('/index.php', '', $url);

		$urlparts = parse_url($url);
		if (empty($urlparts)) {
			return '';
		}

		unset($urlparts['user']);
		unset($urlparts['pass']);
		unset($urlparts['query']);
		unset($urlparts['fragment']);
		return (string)Uri::fromParts($urlparts);
	}

	/**
	 * Detect server data (type, protocol, version number, ...)
	 * The detected data is then updated or inserted in the gserver table.
	 *
	 * @param string  $url           URL of the given server
	 * @param string  $network       Network value that is used, when detection failed
	 * @param boolean $only_nodeinfo Only use nodeinfo for server detection
	 *
	 * @return boolean 'true' if server could be detected
	 */
	public static function detect(string $url, string $network = '', bool $only_nodeinfo = false): bool
	{
		Logger::info('Detect server type', ['server' => $url]);

		$original_url = $url;

		// Remove URL content that is not supposed to exist for a server url
		$url = rtrim(self::cleanURL($url), '/');
		if (empty($url)) {
			Logger::notice('Empty URL.');
			return false;
		}

		// If the URL missmatches, then we mark the old entry as failure
		if (!Strings::compareLink($url, $original_url)) {
			self::setFailure($original_url);
			if (!self::getID($url, true)) {
				self::detect($url, $network, $only_nodeinfo);
			}
			return false;
		}

		$valid_url = Network::isUrlValid($url);
		if (!$valid_url) {
			self::setFailure($url);
			return false;
		} else {
			$valid_url = rtrim($valid_url, '/');
		}

		if (!Strings::compareLink($url, $valid_url)) {
			// We only follow redirects when the path stays the same or the target url has no path.
			// Some systems have got redirects on their landing page to a single account page. This check handles it.
			if (((parse_url($url, PHP_URL_HOST) != parse_url($valid_url, PHP_URL_HOST)) && (parse_url($url, PHP_URL_PATH) == parse_url($valid_url, PHP_URL_PATH))) ||
				(((parse_url($url, PHP_URL_HOST) != parse_url($valid_url, PHP_URL_HOST)) || (parse_url($url, PHP_URL_PATH) != parse_url($valid_url, PHP_URL_PATH))) && empty(parse_url($valid_url, PHP_URL_PATH)))) {
				Logger::debug('Found redirect. Mark old entry as failure', ['old' => $url, 'new' => $valid_url]);
				self::setFailure($url);
				if (!self::getID($valid_url, true)) {
					self::detect($valid_url, $network, $only_nodeinfo);
				}
				return false;
			}

			if ((parse_url($url, PHP_URL_HOST) != parse_url($valid_url, PHP_URL_HOST)) && (parse_url($url, PHP_URL_PATH) != parse_url($valid_url, PHP_URL_PATH)) &&
				(parse_url($url, PHP_URL_PATH) == '')) {
				Logger::debug('Found redirect. Mark old entry as failure and redirect to the basepath.', ['old' => $url, 'new' => $valid_url]);
				$parts = parse_url($valid_url);
				unset($parts['path']);
				$valid_url = (string)Uri::fromParts($parts);

				self::setFailure($url);
				if (!self::getID($valid_url, true)) {
					self::detect($valid_url, $network, $only_nodeinfo);
				}
				return false;
			}
			Logger::debug('Found redirect, but ignore it.', ['old' => $url, 'new' => $valid_url]);
		}

		if ((parse_url($url, PHP_URL_HOST) == parse_url($valid_url, PHP_URL_HOST)) &&
			(parse_url($url, PHP_URL_PATH) == parse_url($valid_url, PHP_URL_PATH)) &&
			(parse_url($url, PHP_URL_SCHEME) != parse_url($valid_url, PHP_URL_SCHEME))) {
			$url = $valid_url;
		}

		$in_webroot = empty(parse_url($url, PHP_URL_PATH));

		// When a nodeinfo is present, we don't need to dig further
		$curlResult = DI::httpClient()->get($url . '/.well-known/x-nodeinfo2', HttpClientAccept::JSON);
		if ($curlResult->isTimeout()) {
			self::setFailure($url);
			return false;
		}

		$serverdata = self::parseNodeinfo210($curlResult);
		if (empty($serverdata)) {
			$curlResult = DI::httpClient()->get($url . '/.well-known/nodeinfo', HttpClientAccept::JSON);
			$serverdata = self::fetchNodeinfo($url, $curlResult);
		}

		if ($only_nodeinfo && empty($serverdata)) {
			Logger::info('Invalid nodeinfo in nodeinfo-mode, server is marked as failure', ['url' => $url]);
			self::setFailure($url);
			return false;
		} elseif (empty($serverdata)) {
			$serverdata = ['detection-method' => self::DETECT_MANUAL, 'network' => Protocol::PHANTOM, 'platform' => '', 'version' => '', 'site_name' => '', 'info' => ''];
		}

		// When there is no Nodeinfo, then use some protocol specific endpoints
		if ($serverdata['network'] == Protocol::PHANTOM) {
			if ($in_webroot) {
				// Fetch the landing page, possibly it reveals some data
				$accept = 'application/activity+json,application/ld+json,application/json,*/*;q=0.9';
				$curlResult = DI::httpClient()->get($url, $accept);
				if (!$curlResult->isSuccess() && $curlResult->getReturnCode() == '406') {
					$curlResult = DI::httpClient()->get($url, HttpClientAccept::HTML);
					$html_fetched = true;
				} else {
					$html_fetched = false;
				}

				if ($curlResult->isSuccess()) {
					$json = json_decode($curlResult->getBody(), true);
					if (!empty($json) && is_array($json)) {
						$data = self::fetchDataFromSystemActor($json, $serverdata);
						$serverdata = $data['server'];
						$systemactor = $data['actor'];
						if (!$html_fetched && !in_array($serverdata['detection-method'], [self::DETECT_SYSTEM_ACTOR, self::DETECT_AP_COLLECTION])) {
							$curlResult = DI::httpClient()->get($url, HttpClientAccept::HTML);
						}
					} elseif (!$html_fetched && (strlen($curlResult->getBody()) < 1000)) {
						$curlResult = DI::httpClient()->get($url, HttpClientAccept::HTML);
					}

					if ($serverdata['detection-method'] != self::DETECT_SYSTEM_ACTOR) {
						$serverdata = self::analyseRootHeader($curlResult, $serverdata);
						$serverdata = self::analyseRootBody($curlResult, $serverdata);
					}
				}

				if (!$curlResult->isSuccess() || empty($curlResult->getBody())) {
					self::setFailure($url);
					return false;
				}

				if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
					$serverdata = self::detectMastodonAlikes($url, $serverdata);
				}
			}

			// All following checks are done for systems that always have got a "host-meta" endpoint.
			// With this check we don't have to waste time and ressources for dead systems.
			// Also this hopefully prevents us from receiving abuse messages.
			if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
				$validHostMeta = self::validHostMeta($url);
			} else {
				$validHostMeta = false;
			}

			if ($validHostMeta) {
				if (in_array($serverdata['detection-method'], [self::DETECT_MANUAL, self::DETECT_HEADER, self::DETECT_BODY])) {
					$serverdata['detection-method'] = self::DETECT_HOST_META;
				}

				if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
					$serverdata = self::detectFriendica($url, $serverdata);
				}

				// The following systems have to be installed in the root directory.
				if ($in_webroot) {
					// the 'siteinfo.json' is some specific endpoint of Hubzilla and Red
					if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
						$serverdata = self::fetchSiteinfo($url, $serverdata);
					}

					// The 'siteinfo.json' doesn't seem to be present on older Hubzilla installations, so we check other endpoints as well
					if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
						$serverdata = self::detectHubzilla($url, $serverdata);
					}

					if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
						$serverdata = self::detectPeertube($url, $serverdata);
					}

					if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
						$serverdata = self::detectGNUSocial($url, $serverdata);
					}
				}
			} elseif (in_array($serverdata['platform'], ['friendica', 'friendika']) && in_array($serverdata['detection-method'], array_merge(self::DETECT_UNSPECIFIC, [self::DETECT_SYSTEM_ACTOR]))) {
				$serverdata = self::detectFriendica($url, $serverdata);
			}

			if (($serverdata['network'] == Protocol::PHANTOM) || in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
				$serverdata = self::detectNextcloud($url, $serverdata, $validHostMeta);
			}

			// When nodeinfo isn't present, we use the older 'statistics.json' endpoint
			// Since this endpoint is only rarely used, we query it at a later time
			if (in_array($serverdata['detection-method'], array_merge(self::DETECT_UNSPECIFIC, [self::DETECT_FRIENDICA, self::DETECT_CONFIG_JSON]))) {
				$serverdata = self::fetchStatistics($url, $serverdata);
			}
		}

		// When we hadn't been able to detect the network type, we use the hint from the parameter
		if (($serverdata['network'] == Protocol::PHANTOM) && !empty($network)) {
			$serverdata['network'] = $network;
		}

		// Most servers aren't installed in a subdirectory, so we declare this entry as failed
		if (($serverdata['network'] == Protocol::PHANTOM) && !empty(parse_url($url, PHP_URL_PATH)) && in_array($serverdata['detection-method'], [self::DETECT_MANUAL])) {
			self::setFailure($url);
			return false;
		}

		$serverdata['url'] = $url;
		$serverdata['nurl'] = Strings::normaliseLink($url);

		// We have to prevent an endless loop here.
		// When a server is new, then there is no gserver entry yet.
		// But in "detectNetworkViaContacts" it could happen that a contact is updated,
		// and this can call this function here as well.
		if (self::getID($url, true) && (in_array($serverdata['network'], [Protocol::PHANTOM, Protocol::FEED]) ||
			in_array($serverdata['detection-method'], [self::DETECT_MANUAL, self::DETECT_HEADER, self::DETECT_BODY, self::DETECT_HOST_META]))) {
			$serverdata = self::detectNetworkViaContacts($url, $serverdata);
		}

		// Detect the directory type
		$serverdata['directory-type'] = self::DT_NONE;

		if (in_array($serverdata['network'], Protocol::FEDERATED)) {
			$serverdata = self::checkMastodonDirectory($url, $serverdata);

			if ($serverdata['directory-type'] == self::DT_NONE) {
				$serverdata = self::checkPoCo($url, $serverdata);
			}
		}

		if ($serverdata['network'] == Protocol::ACTIVITYPUB) {
			$serverdata = self::fetchWeeklyUsage($url, $serverdata);
		}

		$serverdata['registered-users'] = $serverdata['registered-users'] ?? 0;

		// Numbers above a reasonable value (10 millions) are ignored
		if ($serverdata['registered-users'] > 10000000) {
			$serverdata['registered-users'] = 0;
		}

		// On an active server there has to be at least a single user
		if (!in_array($serverdata['network'], [Protocol::PHANTOM, Protocol::FEED]) && ($serverdata['registered-users'] <= 0)) {
			$serverdata['registered-users'] = 1;
		} elseif (in_array($serverdata['network'], [Protocol::PHANTOM, Protocol::FEED])) {
			$serverdata['registered-users'] = 0;
		}

		$serverdata['next_contact'] = self::getNextUpdateDate(true, '', '', in_array($serverdata['network'], [Protocol::PHANTOM, Protocol::FEED]));

		$serverdata['last_contact'] = DateTimeFormat::utcNow();
		$serverdata['failed'] = false;

		$gserver = DBA::selectFirst('gserver', ['network'], ['nurl' => Strings::normaliseLink($url)]);
		if (!DBA::isResult($gserver)) {
			$serverdata['created'] = DateTimeFormat::utcNow();
			$ret = self::insert($serverdata);
			$id = DBA::lastInsertId();
		} else {
			$ret = self::update($serverdata, ['nurl' => $serverdata['nurl']]);
			$gserver = DBA::selectFirst('gserver', ['id'], ['nurl' => $serverdata['nurl']]);
			if (DBA::isResult($gserver)) {
				$id = $gserver['id'];
			}
		}

		// Count the number of known contacts from this server
		if (!empty($id) && !in_array($serverdata['network'], [Protocol::PHANTOM, Protocol::FEED])) {
			$apcontacts = DBA::count('apcontact', ['gsid' => $id]);
			$contacts = DBA::count('contact', ['uid' => 0, 'gsid' => $id, 'failed' => false]);
			$max_users = max($apcontacts, $contacts);
			if ($max_users > $serverdata['registered-users']) {
				Logger::info('Update registered users', ['id' => $id, 'url' => $serverdata['nurl'], 'registered-users' => $max_users]);
				self::update(['registered-users' => $max_users], ['id' => $id]);
			}

			if (empty($serverdata['active-month-users'])) {
				$contacts = DBA::count('contact', ["`uid` = ? AND `gsid` = ? AND NOT `failed` AND `last-item` > ?", 0, $id, DateTimeFormat::utc('now - 30 days')]);
				if ($contacts > 0) {
					Logger::info('Update monthly users', ['id' => $id, 'url' => $serverdata['nurl'], 'monthly-users' => $contacts]);
					self::update(['active-month-users' => $contacts], ['id' => $id]);
				}
			}

			if (empty($serverdata['active-halfyear-users'])) {
				$contacts = DBA::count('contact', ["`uid` = ? AND `gsid` = ? AND NOT `failed` AND `last-item` > ?", 0, $id, DateTimeFormat::utc('now - 180 days')]);
				if ($contacts > 0) {
					Logger::info('Update halfyear users', ['id' => $id, 'url' => $serverdata['nurl'], 'halfyear-users' => $contacts]);
					self::update(['active-halfyear-users' => $contacts], ['id' => $id]);
				}
			}
		}

		if (in_array($serverdata['network'], [Protocol::DFRN, Protocol::DIASPORA])) {
			self::discoverRelay($url);
		}

		if (!empty($systemactor)) {
			$contact = Contact::getByURL($systemactor, true, ['gsid', 'baseurl', 'id', 'network', 'url', 'name']);
			Logger::debug('Fetched system actor',  ['url' => $url, 'gsid' => $id, 'contact' => $contact]);
		}

		return $ret;
	}

	/**
	 * Fetch relay data from a given server url
	 *
	 * @param string $server_url address of the server
	 *
	 * @return void
	 *
	 * @throws \Friendica\Network\HTTPException\InternalServerErrorException
	 */
	private static function discoverRelay(string $server_url)
	{
		Logger::info('Discover relay data', ['server' => $server_url]);

		$curlResult = DI::httpClient()->get($server_url . '/.well-known/x-social-relay', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess()) {
			return;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (!is_array($data)) {
			return;
		}

		// Sanitize incoming data, see https://github.com/friendica/friendica/issues/8565
		$data['subscribe'] = (bool)$data['subscribe'] ?? false;

		if (!$data['subscribe'] || empty($data['scope']) || !in_array(strtolower($data['scope']), ['all', 'tags'])) {
			$data['scope'] = '';
			$data['subscribe'] = false;
			$data['tags'] = [];
		}

		$gserver = DBA::selectFirst('gserver', ['id', 'url', 'network', 'relay-subscribe', 'relay-scope'], ['nurl' => Strings::normaliseLink($server_url)]);
		if (!DBA::isResult($gserver)) {
			return;
		}

		if (($gserver['relay-subscribe'] != $data['subscribe']) || ($gserver['relay-scope'] != $data['scope'])) {
			$fields = ['relay-subscribe' => $data['subscribe'], 'relay-scope' => $data['scope']];
			self::update($fields, ['id' => $gserver['id']]);
		}

		DBA::delete('gserver-tag', ['gserver-id' => $gserver['id']]);

		if ($data['scope'] == 'tags') {
			// Avoid duplicates
			$tags = [];
			foreach ($data['tags'] as $tag) {
				$tag = mb_strtolower($tag);
				if (strlen($tag) < 100) {
					$tags[$tag] = $tag;
				}
			}

			foreach ($tags as $tag) {
				DBA::insert('gserver-tag', ['gserver-id' => $gserver['id'], 'tag' => $tag], Database::INSERT_IGNORE);
			}
		}

		// Create or update the relay contact
		$fields = [];
		if (isset($data['protocols'])) {
			if (isset($data['protocols']['diaspora'])) {
				$fields['network'] = Protocol::DIASPORA;

				if (isset($data['protocols']['diaspora']['receive'])) {
					$fields['batch'] = $data['protocols']['diaspora']['receive'];
				} elseif (is_string($data['protocols']['diaspora'])) {
					$fields['batch'] = $data['protocols']['diaspora'];
				}
			}

			if (isset($data['protocols']['dfrn'])) {
				$fields['network'] = Protocol::DFRN;

				if (isset($data['protocols']['dfrn']['receive'])) {
					$fields['batch'] = $data['protocols']['dfrn']['receive'];
				} elseif (is_string($data['protocols']['dfrn'])) {
					$fields['batch'] = $data['protocols']['dfrn'];
				}
			}

			if (isset($data['protocols']['activitypub'])) {
				$fields['network'] = Protocol::ACTIVITYPUB;

				if (!empty($data['protocols']['activitypub']['actor'])) {
					$fields['url'] = $data['protocols']['activitypub']['actor'];
				}
				if (!empty($data['protocols']['activitypub']['receive'])) {
					$fields['batch'] = $data['protocols']['activitypub']['receive'];
				}
			}
		}

		Logger::info('Discovery ended', ['server' => $server_url, 'data' => $fields]);

		Relay::updateContact($gserver, $fields);
	}

	/**
	 * Fetch server data from '/statistics.json' on the given server
	 *
	 * @param string $url URL of the given server
	 *
	 * @return array server data
	 */
	private static function fetchStatistics(string $url, array $serverdata): array
	{
		$curlResult = DI::httpClient()->get($url . '/statistics.json', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess()) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		// Some AP enabled systems return activity data that we don't expect here.
		if (strpos($curlResult->getContentType(), 'application/activity+json') !== false) {
			return $serverdata;
		}

		$valid = false;
		$old_serverdata = $serverdata;

		$serverdata['detection-method'] = self::DETECT_STATISTICS_JSON;

		if (!empty($data['version'])) {
			$valid = true;
			$serverdata['version'] = $data['version'];
			// Version numbers on statistics.json are presented with additional info, e.g.:
			// 0.6.3.0-p1702cc1c, 0.6.99.0-p1b9ab160 or 3.4.3-2-1191.
			$serverdata['version'] = preg_replace('=(.+)-(.{4,})=ism', '$1', $serverdata['version']);
		}

		if (!empty($data['name'])) {
			$valid = true;
			$serverdata['site_name'] = $data['name'];
		}

		if (!empty($data['network'])) {
			$valid = true;
			$serverdata['platform'] = strtolower($data['network']);

			if ($serverdata['platform'] == 'diaspora') {
				$serverdata['network'] = Protocol::DIASPORA;
			} elseif ($serverdata['platform'] == 'friendica') {
				$serverdata['network'] = Protocol::DFRN;
			} elseif ($serverdata['platform'] == 'hubzilla') {
				$serverdata['network'] = Protocol::ZOT;
			} elseif ($serverdata['platform'] == 'redmatrix') {
				$serverdata['network'] = Protocol::ZOT;
			}
		}

		if (!empty($data['total_users'])) {
			$valid = true;
			$serverdata['registered-users'] = max($data['total_users'], 1);
		}

		if (!empty($data['active_users_monthly'])) {
			$valid = true;
			$serverdata['active-month-users'] = max($data['active_users_monthly'], 0);
		}

		if (!empty($data['active_users_halfyear'])) {
			$valid = true;
			$serverdata['active-halfyear-users'] = max($data['active_users_halfyear'], 0);
		}

		if (!empty($data['local_posts'])) {
			$valid = true;
			$serverdata['local-posts'] = max($data['local_posts'], 0);
		}

		if (!empty($data['registrations_open'])) {
			$serverdata['register_policy'] = Register::OPEN;
		} else {
			$serverdata['register_policy'] = Register::CLOSED;
		}

		if (!$valid) {
			return $old_serverdata;
		}

		return $serverdata;
	}

	/**
	 * Detect server type by using the nodeinfo data
	 *
	 * @param string                  $url        address of the server
	 * @param ICanHandleHttpResponses $httpResult
	 *
	 * @return array Server data
	 *
	 * @throws \Friendica\Network\HTTPException\InternalServerErrorException
	 */
	private static function fetchNodeinfo(string $url, ICanHandleHttpResponses $httpResult): array
	{
		if (!$httpResult->isSuccess()) {
			return [];
		}

		$nodeinfo = json_decode($httpResult->getBody(), true);

		if (!is_array($nodeinfo) || empty($nodeinfo['links'])) {
			return [];
		}

		$nodeinfo1_url = '';
		$nodeinfo2_url = '';

		foreach ($nodeinfo['links'] as $link) {
			if (!is_array($link) || empty($link['rel']) || empty($link['href'])) {
				Logger::info('Invalid nodeinfo format', ['url' => $url]);
				continue;
			}
			if ($link['rel'] == 'http://nodeinfo.diaspora.software/ns/schema/1.0') {
				$nodeinfo1_url = $link['href'];
			} elseif ($link['rel'] == 'http://nodeinfo.diaspora.software/ns/schema/2.0') {
				$nodeinfo2_url = $link['href'];
			}
		}

		if ($nodeinfo1_url . $nodeinfo2_url == '') {
			return [];
		}

		$server = [];

		if (!empty($nodeinfo2_url)) {
			$server = self::parseNodeinfo2($nodeinfo2_url);
		}

		if (empty($server) && !empty($nodeinfo1_url)) {
			$server = self::parseNodeinfo1($nodeinfo1_url);
		}

		return $server;
	}

	/**
	 * Parses Nodeinfo 1
	 *
	 * @param string $nodeinfo_url address of the nodeinfo path
	 *
	 * @return array Server data
	 *
	 * @throws \Friendica\Network\HTTPException\InternalServerErrorException
	 */
	private static function parseNodeinfo1(string $nodeinfo_url): array
	{
		$curlResult = DI::httpClient()->get($nodeinfo_url, HttpClientAccept::JSON);
		if (!$curlResult->isSuccess()) {
			return [];
		}

		$nodeinfo = json_decode($curlResult->getBody(), true);

		if (!is_array($nodeinfo)) {
			return [];
		}

		$server = ['detection-method' => self::DETECT_NODEINFO_1,
			'register_policy' => Register::CLOSED];

		if (!empty($nodeinfo['openRegistrations'])) {
			$server['register_policy'] = Register::OPEN;
		}

		if (is_array($nodeinfo['software'])) {
			if (!empty($nodeinfo['software']['name'])) {
				$server['platform'] = strtolower($nodeinfo['software']['name']);
			}

			if (!empty($nodeinfo['software']['version'])) {
				$server['version'] = $nodeinfo['software']['version'];
				// Version numbers on Nodeinfo are presented with additional info, e.g.:
				// 0.6.3.0-p1702cc1c, 0.6.99.0-p1b9ab160 or 3.4.3-2-1191.
				$server['version'] = preg_replace('=(.+)-(.{4,})=ism', '$1', $server['version']);
			}
		}

		if (!empty($nodeinfo['metadata']['nodeName'])) {
			$server['site_name'] = $nodeinfo['metadata']['nodeName'];
		}

		if (!empty($nodeinfo['usage']['users']['total'])) {
			$server['registered-users'] = max($nodeinfo['usage']['users']['total'], 1);
		}

		if (!empty($nodeinfo['usage']['users']['activeMonth'])) {
			$server['active-month-users'] = max($nodeinfo['usage']['users']['activeMonth'], 0);
		}

		if (!empty($nodeinfo['usage']['users']['activeHalfyear'])) {
			$server['active-halfyear-users'] = max($nodeinfo['usage']['users']['activeHalfyear'], 0);
		}

		if (!empty($nodeinfo['usage']['localPosts'])) {
			$server['local-posts'] = max($nodeinfo['usage']['localPosts'], 0);
		}

		if (!empty($nodeinfo['usage']['localComments'])) {
			$server['local-comments'] = max($nodeinfo['usage']['localComments'], 0);
		}

		if (!empty($nodeinfo['protocols']['inbound']) && is_array($nodeinfo['protocols']['inbound'])) {
			$protocols = [];
			foreach ($nodeinfo['protocols']['inbound'] as $protocol) {
				$protocols[$protocol] = true;
			}

			if (!empty($protocols['friendica'])) {
				$server['network'] = Protocol::DFRN;
			} elseif (!empty($protocols['activitypub'])) {
				$server['network'] = Protocol::ACTIVITYPUB;
			} elseif (!empty($protocols['diaspora'])) {
				$server['network'] = Protocol::DIASPORA;
			} elseif (!empty($protocols['ostatus'])) {
				$server['network'] = Protocol::OSTATUS;
			} elseif (!empty($protocols['gnusocial'])) {
				$server['network'] = Protocol::OSTATUS;
			} elseif (!empty($protocols['zot'])) {
				$server['network'] = Protocol::ZOT;
			}
		}

		if (empty($server)) {
			return [];
		}

		if (empty($server['network'])) {
			$server['network'] = Protocol::PHANTOM;
		}

		return $server;
	}

	/**
	 * Parses Nodeinfo 2
	 *
	 * @see https://git.feneas.org/jaywink/nodeinfo2
	 *
	 * @param string $nodeinfo_url address of the nodeinfo path
	 *
	 * @return array Server data
	 *
	 * @throws \Friendica\Network\HTTPException\InternalServerErrorException
	 */
	private static function parseNodeinfo2(string $nodeinfo_url): array
	{
		$curlResult = DI::httpClient()->get($nodeinfo_url, HttpClientAccept::JSON);
		if (!$curlResult->isSuccess()) {
			return [];
		}

		$nodeinfo = json_decode($curlResult->getBody(), true);
		if (!is_array($nodeinfo)) {
			return [];
		}

		$server = [
			'detection-method' => self::DETECT_NODEINFO_2,
			'register_policy' => Register::CLOSED,
			'platform' => 'unknown',
		];

		if (!empty($nodeinfo['openRegistrations'])) {
			$server['register_policy'] = Register::OPEN;
		}

		if (!empty($nodeinfo['software'])) {
			if (isset($nodeinfo['software']['name'])) {
				$server['platform'] = strtolower($nodeinfo['software']['name']);
			}

			if (!empty($nodeinfo['software']['version']) && isset($server['platform'])) {
				$server['version'] = $nodeinfo['software']['version'];
				// Version numbers on Nodeinfo are presented with additional info, e.g.:
				// 0.6.3.0-p1702cc1c, 0.6.99.0-p1b9ab160 or 3.4.3-2-1191.
				$server['version'] = preg_replace('=(.+)-(.{4,})=ism', '$1', $server['version']);

				// qoto advertises itself as Mastodon
				if (($server['platform'] == 'mastodon') && substr($nodeinfo['software']['version'], -5) == '-qoto') {
					$server['platform'] = 'qoto';
				}
			}
		}

		if (!empty($nodeinfo['metadata']['nodeName'])) {
			$server['site_name'] = $nodeinfo['metadata']['nodeName'];
		}

		if (!empty($nodeinfo['usage']['users']['total'])) {
			$server['registered-users'] = max($nodeinfo['usage']['users']['total'], 1);
		}

		if (!empty($nodeinfo['usage']['users']['activeMonth'])) {
			$server['active-month-users'] = max($nodeinfo['usage']['users']['activeMonth'], 0);
		}

		if (!empty($nodeinfo['usage']['users']['activeHalfyear'])) {
			$server['active-halfyear-users'] = max($nodeinfo['usage']['users']['activeHalfyear'], 0);
		}

		if (!empty($nodeinfo['usage']['localPosts'])) {
			$server['local-posts'] = max($nodeinfo['usage']['localPosts'], 0);
		}

		if (!empty($nodeinfo['usage']['localComments'])) {
			$server['local-comments'] = max($nodeinfo['usage']['localComments'], 0);
		}

		if (!empty($nodeinfo['protocols'])) {
			$protocols = [];
			foreach ($nodeinfo['protocols'] as $protocol) {
				if (is_string($protocol)) {
					$protocols[$protocol] = true;
				}
			}

			if (!empty($protocols['dfrn'])) {
				$server['network'] = Protocol::DFRN;
			} elseif (!empty($protocols['activitypub'])) {
				$server['network'] = Protocol::ACTIVITYPUB;
			} elseif (!empty($protocols['diaspora'])) {
				$server['network'] = Protocol::DIASPORA;
			} elseif (!empty($protocols['ostatus'])) {
				$server['network'] = Protocol::OSTATUS;
			} elseif (!empty($protocols['gnusocial'])) {
				$server['network'] = Protocol::OSTATUS;
			} elseif (!empty($protocols['zot'])) {
				$server['network'] = Protocol::ZOT;
			}
		}

		if (empty($server)) {
			return [];
		}

		if (empty($server['network'])) {
			$server['network'] = Protocol::PHANTOM;
		}

		return $server;
	}

	/**
	 * Parses NodeInfo2 protocol 1.0
	 *
	 * @see https://github.com/jaywink/nodeinfo2/blob/master/PROTOCOL.md
	 *
	 * @param string $nodeinfo_url address of the nodeinfo path
	 *
	 * @return array Server data
	 *
	 * @throws \Friendica\Network\HTTPException\InternalServerErrorException
	 */
	private static function parseNodeinfo210(ICanHandleHttpResponses $httpResult): array
	{
		if (!$httpResult->isSuccess()) {
			return [];
		}

		$nodeinfo = json_decode($httpResult->getBody(), true);

		if (!is_array($nodeinfo)) {
			return [];
		}

		$server = ['detection-method' => self::DETECT_NODEINFO_210,
			'register_policy' => Register::CLOSED];

		if (!empty($nodeinfo['openRegistrations'])) {
			$server['register_policy'] = Register::OPEN;
		}

		if (!empty($nodeinfo['server'])) {
			if (!empty($nodeinfo['server']['software'])) {
				$server['platform'] = strtolower($nodeinfo['server']['software']);
			}

			if (!empty($nodeinfo['server']['version'])) {
				$server['version'] = $nodeinfo['server']['version'];
				// Version numbers on Nodeinfo are presented with additional info, e.g.:
				// 0.6.3.0-p1702cc1c, 0.6.99.0-p1b9ab160 or 3.4.3-2-1191.
				$server['version'] = preg_replace('=(.+)-(.{4,})=ism', '$1', $server['version']);
			}

			if (!empty($nodeinfo['server']['name'])) {
				$server['site_name'] = $nodeinfo['server']['name'];
			}
		}

		if (!empty($nodeinfo['usage']['users']['total'])) {
			$server['registered-users'] = max($nodeinfo['usage']['users']['total'], 1);
		}

		if (!empty($nodeinfo['usage']['users']['activeMonth'])) {
			$server['active-month-users'] = max($nodeinfo['usage']['users']['activeMonth'], 0);
		}

		if (!empty($nodeinfo['usage']['users']['activeHalfyear'])) {
			$server['active-halfyear-users'] = max($nodeinfo['usage']['users']['activeHalfyear'], 0);
		}

		if (!empty($nodeinfo['usage']['localPosts'])) {
			$server['local-posts'] = max($nodeinfo['usage']['localPosts'], 0);
		}

		if (!empty($nodeinfo['usage']['localComments'])) {
			$server['local-comments'] = max($nodeinfo['usage']['localComments'], 0);
		}

		if (!empty($nodeinfo['protocols'])) {
			$protocols = [];
			foreach ($nodeinfo['protocols'] as $protocol) {
				if (is_string($protocol)) {
					$protocols[$protocol] = true;
				}
			}

			if (!empty($protocols['dfrn'])) {
				$server['network'] = Protocol::DFRN;
			} elseif (!empty($protocols['activitypub'])) {
				$server['network'] = Protocol::ACTIVITYPUB;
			} elseif (!empty($protocols['diaspora'])) {
				$server['network'] = Protocol::DIASPORA;
			} elseif (!empty($protocols['ostatus'])) {
				$server['network'] = Protocol::OSTATUS;
			} elseif (!empty($protocols['gnusocial'])) {
				$server['network'] = Protocol::OSTATUS;
			} elseif (!empty($protocols['zot'])) {
				$server['network'] = Protocol::ZOT;
			}
		}

		if (empty($server) || empty($server['platform'])) {
			return [];
		}

		if (empty($server['network'])) {
			$server['network'] = Protocol::PHANTOM;
		}

		return $server;
	}

	/**
	 * Fetch server information from a 'siteinfo.json' file on the given server
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function fetchSiteinfo(string $url, array $serverdata): array
	{
		$curlResult = DI::httpClient()->get($url . '/siteinfo.json', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess()) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		if (in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
			$serverdata['detection-method'] = self::DETECT_SITEINFO_JSON;
		}

		if (!empty($data['url'])) {
			$serverdata['platform'] = strtolower($data['platform']);
			$serverdata['version'] = $data['version'] ?? 'N/A';
		}

		if (!empty($data['plugins'])) {
			if (in_array('pubcrawl', $data['plugins'])) {
				$serverdata['network'] = Protocol::ACTIVITYPUB;
			} elseif (in_array('diaspora', $data['plugins'])) {
				$serverdata['network'] = Protocol::DIASPORA;
			} elseif (in_array('gnusoc', $data['plugins'])) {
				$serverdata['network'] = Protocol::OSTATUS;
			} else {
				$serverdata['network'] = Protocol::ZOT;
			}
		}

		if (!empty($data['site_name'])) {
			$serverdata['site_name'] = $data['site_name'];
		}

		if (!empty($data['channels_total'])) {
			$serverdata['registered-users'] = max($data['channels_total'], 1);
		}

		if (!empty($data['channels_active_monthly'])) {
			$serverdata['active-month-users'] = max($data['channels_active_monthly'], 0);
		}

		if (!empty($data['channels_active_halfyear'])) {
			$serverdata['active-halfyear-users'] = max($data['channels_active_halfyear'], 0);
		}

		if (!empty($data['local_posts'])) {
			$serverdata['local-posts'] = max($data['local_posts'], 0);
		}

		if (!empty($data['local_comments'])) {
			$serverdata['local-comments'] = max($data['local_comments'], 0);
		}

		if (!empty($data['register_policy'])) {
			switch ($data['register_policy']) {
				case 'REGISTER_OPEN':
					$serverdata['register_policy'] = Register::OPEN;
					break;

				case 'REGISTER_APPROVE':
					$serverdata['register_policy'] = Register::APPROVE;
					break;

				case 'REGISTER_CLOSED':
				default:
					$serverdata['register_policy'] = Register::CLOSED;
					break;
			}
		}

		return $serverdata;
	}

	/**
	 * Fetches server data via an ActivityPub account with url of that server
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 *
	 * @throws Exception
	 */
	private static function fetchDataFromSystemActor(array $data, array $serverdata): array
	{
		if (empty($data)) {
			return ['server' => $serverdata, 'actor' => ''];
		}

		$actor = JsonLD::compact($data, false);
		if (in_array(JsonLD::fetchElement($actor, '@type'), ActivityPub\Receiver::ACCOUNT_TYPES)) {
			$serverdata['network'] = Protocol::ACTIVITYPUB;
			$serverdata['site_name'] = JsonLD::fetchElement($actor, 'as:name', '@value');
			$serverdata['info'] = JsonLD::fetchElement($actor, 'as:summary', '@value');
			if (!empty($actor['as:generator'])) {
				$generator = explode(' ', JsonLD::fetchElement($actor['as:generator'], 'as:name', '@value'));
				$serverdata['platform'] = strtolower(array_shift($generator));
				$serverdata['detection-method'] = self::DETECT_SYSTEM_ACTOR;
			} else {
				$serverdata['detection-method'] = self::DETECT_AP_ACTOR;
			}
			return ['server' => $serverdata, 'actor' => $actor['@id']];
		} elseif ((JsonLD::fetchElement($actor, '@type') == 'as:Collection')) {
			// By now only Ktistec seems to provide collections this way
			$serverdata['platform'] = 'ktistec';
			$serverdata['network'] = Protocol::ACTIVITYPUB;
			$serverdata['detection-method'] = self::DETECT_AP_COLLECTION;

			$actors = JsonLD::fetchElementArray($actor, 'as:items');
			if (!empty($actors) && !empty($actors[0]['@id'])) {
				$actor_url = $actor['@id'] . $actors[0]['@id'];
			} else {
				$actor_url = '';
			}

			return ['server' => $serverdata, 'actor' => $actor_url];
		}
		return ['server' => $serverdata, 'actor' => ''];
	}

	/**
	 * Checks if the server contains a valid host meta file
	 *
	 * @param string $url URL of the given server
	 *
	 * @return boolean 'true' if the server seems to be vital
	 */
	private static function validHostMeta(string $url): bool
	{
		$xrd_timeout = DI::config()->get('system', 'xrd_timeout');
		$curlResult = DI::httpClient()->get($url . Probe::HOST_META, HttpClientAccept::XRD_XML, [HttpClientOptions::TIMEOUT => $xrd_timeout]);
		if (!$curlResult->isSuccess()) {
			return false;
		}

		$xrd = XML::parseString($curlResult->getBody(), true);
		if (!is_object($xrd)) {
			return false;
		}

		$elements = XML::elementToArray($xrd);
		if (empty($elements) || empty($elements['xrd']) || empty($elements['xrd']['link'])) {
			return false;
		}

		$valid = false;
		foreach ($elements['xrd']['link'] as $link) {
			// When there is more than a single "link" element, the array looks slightly different
			if (!empty($link['@attributes'])) {
				$link = $link['@attributes'];
			}

			if (empty($link['rel']) || empty($link['template'])) {
				continue;
			}

			if ($link['rel'] == 'lrdd') {
				// When the webfinger host is the same like the system host, it should be ok.
				$valid = (parse_url($url, PHP_URL_HOST) == parse_url($link['template'], PHP_URL_HOST));
			}
		}

		return $valid;
	}

	/**
	 * Detect the network of the given server via their known contacts
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function detectNetworkViaContacts(string $url, array $serverdata): array
	{
		$contacts = [];

		$nurl = Strings::normaliseLink($url);

		$apcontacts = DBA::select('apcontact', ['url'], ['baseurl' => [$url, $nurl]]);
		while ($apcontact = DBA::fetch($apcontacts)) {
			$contacts[Strings::normaliseLink($apcontact['url'])] = $apcontact['url'];
		}
		DBA::close($apcontacts);

		$pcontacts = DBA::select('contact', ['url', 'nurl'], ['uid' => 0, 'baseurl' => [$url, $nurl]]);
		while ($pcontact = DBA::fetch($pcontacts)) {
			$contacts[$pcontact['nurl']] = $pcontact['url'];
		}
		DBA::close($pcontacts);

		if (empty($contacts)) {
			return $serverdata;
		}

		$time = time();
		foreach ($contacts as $contact) {
			// Endlosschleife verhindern wegen gsid!
			$data = Probe::uri($contact);
			if (in_array($data['network'], Protocol::FEDERATED)) {
				$serverdata['network'] = $data['network'];

				if (in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
					$serverdata['detection-method'] = self::DETECT_CONTACTS;
				}
				break;
			} elseif ((time() - $time) > 10) {
				// To reduce the stress on remote systems we probe a maximum of 10 seconds
				break;
			}
		}

		return $serverdata;
	}

	/**
	 * Checks if the given server does have a '/poco' endpoint.
	 * This is used for the 'PortableContact' functionality,
	 * which is used by both Friendica and Hubzilla.
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function checkPoCo(string $url, array $serverdata): array
	{
		$serverdata['poco'] = '';

		$curlResult = DI::httpClient()->get($url . '/poco', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess()) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		if (!empty($data['totalResults'])) {
			$registeredUsers = $serverdata['registered-users'] ?? 0;
			$serverdata['registered-users'] = max($data['totalResults'], $registeredUsers, 1);
			$serverdata['directory-type'] = self::DT_POCO;
			$serverdata['poco'] = $url . '/poco';
		}

		return $serverdata;
	}

	/**
	 * Checks if the given server does have a Mastodon style directory endpoint.
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	public static function checkMastodonDirectory(string $url, array $serverdata): array
	{
		$curlResult = DI::httpClient()->get($url . '/api/v1/directory?limit=1', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess()) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		if (count($data) == 1) {
			$serverdata['directory-type'] = self::DT_MASTODON;
		}

		return $serverdata;
	}

	/**
	 * Detects Peertube via their known endpoint
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function detectPeertube(string $url, array $serverdata): array
	{
		$curlResult = DI::httpClient()->get($url . '/api/v1/config', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess() || ($curlResult->getBody() == '')) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		if (!empty($data['instance']) && !empty($data['serverVersion'])) {
			$serverdata['platform'] = 'peertube';
			$serverdata['version'] = $data['serverVersion'];
			$serverdata['network'] = Protocol::ACTIVITYPUB;

			if (!empty($data['instance']['name'])) {
				$serverdata['site_name'] = $data['instance']['name'];
			}

			if (!empty($data['instance']['shortDescription'])) {
				$serverdata['info'] = $data['instance']['shortDescription'];
			}

			if (!empty($data['signup'])) {
				if (!empty($data['signup']['allowed'])) {
					$serverdata['register_policy'] = Register::OPEN;
				}
			}

			if (in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
				$serverdata['detection-method'] = self::DETECT_V1_CONFIG;
			}
		}

		return $serverdata;
	}

	/**
	 * Detects the version number of a given server when it was a NextCloud installation
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 * @param bool   $validHostMeta
	 *
	 * @return array server data
	 */
	private static function detectNextcloud(string $url, array $serverdata, bool $validHostMeta): array
	{
		$curlResult = DI::httpClient()->get($url . '/status.php', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess() || ($curlResult->getBody() == '')) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		if (!empty($data['version'])) {
			$serverdata['platform'] = 'nextcloud';
			$serverdata['version'] = $data['version'];

			if ($validHostMeta) {
				$serverdata['network'] = Protocol::ACTIVITYPUB;
			}

			if (in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
				$serverdata['detection-method'] = self::DETECT_STATUS_PHP;
			}
		}

		return $serverdata;
	}

	/**
	 * Fetches weekly usage data
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function fetchWeeklyUsage(string $url, array $serverdata): array
	{
		$curlResult = DI::httpClient()->get($url . '/api/v1/instance/activity', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess() || ($curlResult->getBody() == '')) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		$current_week = [];
		foreach ($data as $week) {
			// Use only data from a full week
			if (empty($week['week']) || (time() - $week['week']) < 7 * 24 * 60 * 60) {
				continue;
			}

			// Most likely the data is sorted correctly. But we better are safe than sorry
			if (empty($current_week['week']) || ($current_week['week'] < $week['week'])) {
				$current_week = $week;
			}
		}

		if (!empty($current_week['logins'])) {
			$serverdata['active-week-users'] = max($current_week['logins'], 0);
		}

		return $serverdata;
	}

	/**
	 * Detects data from a given server url if it was a mastodon alike system
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function detectMastodonAlikes(string $url, array $serverdata): array
	{
		$curlResult = DI::httpClient()->get($url . '/api/v1/instance', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess() || ($curlResult->getBody() == '')) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data)) {
			return $serverdata;
		}

		$valid = false;

		if (!empty($data['version'])) {
			$serverdata['platform'] = 'mastodon';
			$serverdata['version'] = $data['version'] ?? '';
			$serverdata['network'] = Protocol::ACTIVITYPUB;
			$valid = true;
		}

		if (!empty($data['title'])) {
			$serverdata['site_name'] = $data['title'];
		}

		if (!empty($data['title']) && empty($serverdata['platform']) && ($serverdata['network'] == Protocol::PHANTOM)) {
			$serverdata['platform'] = 'mastodon';
			$serverdata['network'] = Protocol::ACTIVITYPUB;
			$valid = true;
		}

		if (!empty($data['description'])) {
			$serverdata['info'] = trim($data['description']);
		}

		if (!empty($data['stats']['user_count'])) {
			$serverdata['registered-users'] = max($data['stats']['user_count'], 1);
		}

		if (!empty($serverdata['version']) && preg_match('/.*?\(compatible;\s(.*)\s(.*)\)/ism', $serverdata['version'], $matches)) {
			$serverdata['platform'] = strtolower($matches[1]);
			$serverdata['version'] = $matches[2];
			$valid = true;
		}

		if (!empty($serverdata['version']) && strstr(strtolower($serverdata['version']), 'pleroma')) {
			$serverdata['platform'] = 'pleroma';
			$serverdata['version'] = trim(str_ireplace('pleroma', '', $serverdata['version']));
			$valid = true;
		}

		if (!empty($serverdata['platform']) && strstr($serverdata['platform'], 'pleroma')) {
			$serverdata['version'] = trim(str_ireplace('pleroma', '', $serverdata['platform']));
			$serverdata['platform'] = 'pleroma';
			$valid = true;
		}

		if ($valid && in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
			$serverdata['detection-method'] = self::DETECT_MASTODON_API;
		}

		return $serverdata;
	}

	/**
	 * Detects data from typical Hubzilla endpoints
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function detectHubzilla(string $url, array $serverdata): array
	{
		$curlResult = DI::httpClient()->get($url . '/api/statusnet/config.json', HttpClientAccept::JSON);
		if (!$curlResult->isSuccess() || ($curlResult->getBody() == '')) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data) || empty($data['site'])) {
			return $serverdata;
		}

		if (!empty($data['site']['name'])) {
			$serverdata['site_name'] = $data['site']['name'];
		}

		if (!empty($data['site']['platform'])) {
			$serverdata['platform'] = strtolower($data['site']['platform']['PLATFORM_NAME']);
			$serverdata['version'] = $data['site']['platform']['STD_VERSION'];
			$serverdata['network'] = Protocol::ZOT;
		}

		if (!empty($data['site']['hubzilla'])) {
			$serverdata['platform'] = strtolower($data['site']['hubzilla']['PLATFORM_NAME']);
			$serverdata['version'] = $data['site']['hubzilla']['RED_VERSION'];
			$serverdata['network'] = Protocol::ZOT;
		}

		if (!empty($data['site']['redmatrix'])) {
			if (!empty($data['site']['redmatrix']['PLATFORM_NAME'])) {
				$serverdata['platform'] = strtolower($data['site']['redmatrix']['PLATFORM_NAME']);
			} elseif (!empty($data['site']['redmatrix']['RED_PLATFORM'])) {
				$serverdata['platform'] = strtolower($data['site']['redmatrix']['RED_PLATFORM']);
			}

			$serverdata['version'] = $data['site']['redmatrix']['RED_VERSION'];
			$serverdata['network'] = Protocol::ZOT;
		}

		$private = false;
		$inviteonly = false;
		$closed = false;

		if (!empty($data['site']['closed'])) {
			$closed = self::toBoolean($data['site']['closed']);
		}

		if (!empty($data['site']['private'])) {
			$private = self::toBoolean($data['site']['private']);
		}

		if (!empty($data['site']['inviteonly'])) {
			$inviteonly = self::toBoolean($data['site']['inviteonly']);
		}

		if (!$closed && !$private and $inviteonly) {
			$serverdata['register_policy'] = Register::APPROVE;
		} elseif (!$closed && !$private) {
			$serverdata['register_policy'] = Register::OPEN;
		} else {
			$serverdata['register_policy'] = Register::CLOSED;
		}

		if (($serverdata['network'] != Protocol::PHANTOM) && in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
			$serverdata['detection-method'] = self::DETECT_CONFIG_JSON;
		}

		return $serverdata;
	}

	/**
	 * Converts input value to a boolean value
	 *
	 * @param string|integer $val
	 *
	 * @return boolean
	 */
	private static function toBoolean($val): bool
	{
		if (($val == 'true') || ($val == 1)) {
			return true;
		} elseif (($val == 'false') || ($val == 0)) {
			return false;
		}

		return $val;
	}

	/**
	 * Detect if the URL belongs to a GNU Social server
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function detectGNUSocial(string $url, array $serverdata): array
	{
		// Test for GNU Social
		$curlResult = DI::httpClient()->get($url . '/api/gnusocial/version.json', HttpClientAccept::JSON);
		if ($curlResult->isSuccess() && ($curlResult->getBody() != '{"error":"not implemented"}') &&
			($curlResult->getBody() != '') && (strlen($curlResult->getBody()) < 30)) {
			$serverdata['platform'] = 'gnusocial';
			// Remove junk that some GNU Social servers return
			$serverdata['version'] = str_replace(chr(239) . chr(187) . chr(191), '', $curlResult->getBody());
			$serverdata['version'] = str_replace(["\r", "\n", "\t"], '', $serverdata['version']);
			$serverdata['version'] = trim($serverdata['version'], '"');
			$serverdata['network'] = Protocol::OSTATUS;

			if (in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
				$serverdata['detection-method'] = self::DETECT_GNUSOCIAL;
			}

			return $serverdata;
		}

		// Test for Statusnet
		$curlResult = DI::httpClient()->get($url . '/api/statusnet/version.json', HttpClientAccept::JSON);
		if ($curlResult->isSuccess() && ($curlResult->getBody() != '{"error":"not implemented"}') &&
			($curlResult->getBody() != '') && (strlen($curlResult->getBody()) < 30)) {

			// Remove junk that some GNU Social servers return
			$serverdata['version'] = str_replace(chr(239).chr(187).chr(191), '', $curlResult->getBody());
			$serverdata['version'] = str_replace(["\r", "\n", "\t"], '', $serverdata['version']);
			$serverdata['version'] = trim($serverdata['version'], '"');

			if (!empty($serverdata['version']) && strtolower(substr($serverdata['version'], 0, 7)) == 'pleroma') {
				$serverdata['platform'] = 'pleroma';
				$serverdata['version'] = trim(str_ireplace('pleroma', '', $serverdata['version']));
				$serverdata['network'] = Protocol::ACTIVITYPUB;
			} else {
				$serverdata['platform'] = 'statusnet';
				$serverdata['network'] = Protocol::OSTATUS;
			}

			if (in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
				$serverdata['detection-method'] = self::DETECT_STATUSNET;
			}
		}

		return $serverdata;
	}

	/**
	 * Detect if the URL belongs to a Friendica server
	 *
	 * @param string $url        URL of the given server
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function detectFriendica(string $url, array $serverdata): array
	{
		// There is a bug in some versions of Friendica that will return an ActivityStream actor when the content type "application/json" is requested.
		// Because of this me must not use ACCEPT_JSON here.
		$curlResult = DI::httpClient()->get($url . '/friendica/json');
		if (!$curlResult->isSuccess()) {
			$curlResult = DI::httpClient()->get($url . '/friendika/json');
			$friendika = true;
			$platform = 'Friendika';
		} else {
			$friendika = false;
			$platform = 'Friendica';
		}

		if (!$curlResult->isSuccess()) {
			return $serverdata;
		}

		$data = json_decode($curlResult->getBody(), true);
		if (empty($data) || empty($data['version'])) {
			return $serverdata;
		}

		if (in_array($serverdata['detection-method'], self::DETECT_UNSPECIFIC)) {
			$serverdata['detection-method'] = $friendika ? self::DETECT_FRIENDIKA : self::DETECT_FRIENDICA;
		}

		$serverdata['network'] = Protocol::DFRN;
		$serverdata['version'] = $data['version'];

		if (!empty($data['no_scrape_url'])) {
			$serverdata['noscrape'] = $data['no_scrape_url'];
		}

		if (!empty($data['site_name'])) {
			$serverdata['site_name'] = $data['site_name'];
		}

		if (!empty($data['info'])) {
			$serverdata['info'] = trim($data['info']);
		}

		$register_policy = ($data['register_policy'] ?? '') ?: 'REGISTER_CLOSED';
		switch ($register_policy) {
			case 'REGISTER_OPEN':
				$serverdata['register_policy'] = Register::OPEN;
				break;

			case 'REGISTER_APPROVE':
				$serverdata['register_policy'] = Register::APPROVE;
				break;

			case 'REGISTER_CLOSED':
			case 'REGISTER_INVITATION':
				$serverdata['register_policy'] = Register::CLOSED;
				break;
			default:
				Logger::info('Register policy is invalid', ['policy' => $register_policy, 'server' => $url]);
 				$serverdata['register_policy'] = Register::CLOSED;
				break;
		}

		$serverdata['platform'] = strtolower($data['platform'] ?? $platform);

		return $serverdata;
	}

	/**
	 * Analyses the landing page of a given server for hints about type and system of that server
	 *
	 * @param object $curlResult result of curl execution
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function analyseRootBody($curlResult, array $serverdata): array
	{
		if (empty($curlResult->getBody())) {
			return $serverdata;
		}

		if (file_exists(__DIR__ . '/../../static/platforms.config.php')) {
			require __DIR__ . '/../../static/platforms.config.php';
		} else {
			throw new HTTPException\InternalServerErrorException('Invalid platform file');
		}

		$platforms = array_merge($ap_platforms, $dfrn_platforms, $zap_platforms, $platforms);

		$doc = new DOMDocument();
		@$doc->loadHTML($curlResult->getBody());
		$xpath = new DOMXPath($doc);
		$assigned = false;

		// We can only detect honk via some HTML element on their page
		if ($xpath->query('//div[@id="honksonpage"]')->count() == 1) {
			$serverdata['platform'] = 'honk';
			$serverdata['network'] = Protocol::ACTIVITYPUB;
			$assigned = true;
		}

		$title = trim(XML::getFirstNodeValue($xpath, '//head/title/text()'));
		if (!empty($title)) {
			$serverdata['site_name'] = $title;
		}

		$list = $xpath->query('//meta[@name]');

		foreach ($list as $node) {
			$attr = [];
			if ($node->attributes->length) {
				foreach ($node->attributes as $attribute) {
					$value = trim($attribute->value);
					if (empty($value)) {
						continue;
					}

					$attr[$attribute->name] = $value;
				}

				if (empty($attr['name']) || empty($attr['content'])) {
					continue;
				}
			}

			if ($attr['name'] == 'description') {
				$serverdata['info'] = $attr['content'];
			}

			if (in_array($attr['name'], ['application-name', 'al:android:app_name', 'al:ios:app_name',
				'twitter:app:name:googleplay', 'twitter:app:name:iphone', 'twitter:app:name:ipad', 'generator'])) {
				$platform = str_ireplace(array_keys($platforms), array_values($platforms), $attr['content']);
				$platform = str_replace('/', ' ', $platform);
				$platform_parts = explode(' ', $platform);
				if ((count($platform_parts) >= 2) && in_array(strtolower($platform_parts[0]), array_values($platforms))) {
					$platform = $platform_parts[0];
					$serverdata['version'] = $platform_parts[1];
				}
				if (in_array($platform, array_values($dfrn_platforms))) {
					$serverdata['network'] = Protocol::DFRN;
				} elseif (in_array($platform, array_values($ap_platforms))) {
					$serverdata['network'] = Protocol::ACTIVITYPUB;
				} elseif (in_array($platform, array_values($zap_platforms))) {
					$serverdata['network'] = Protocol::ZOT;
				}
				if (in_array($platform, array_values($platforms))) {
					$serverdata['platform'] = $platform;
					$assigned = true;
				}
			}
		}

		$list = $xpath->query('//meta[@property]');

		foreach ($list as $node) {
			$attr = [];
			if ($node->attributes->length) {
				foreach ($node->attributes as $attribute) {
					$value = trim($attribute->value);
					if (empty($value)) {
						continue;
					}

					$attr[$attribute->name] = $value;
				}

				if (empty($attr['property']) || empty($attr['content'])) {
					continue;
				}
			}

			if ($attr['property'] == 'og:site_name') {
				$serverdata['site_name'] = $attr['content'];
			}

			if ($attr['property'] == 'og:description') {
				$serverdata['info'] = $attr['content'];
			}

			if (in_array($attr['property'], ['og:platform', 'generator'])) {
				if (in_array($attr['content'], array_keys($platforms))) {
					$serverdata['platform'] = $platforms[$attr['content']];
					$assigned = true;
				}

				if (in_array($attr['content'], array_keys($ap_platforms))) {
					$serverdata['network'] = Protocol::ACTIVITYPUB;
				} elseif (in_array($attr['content'], array_values($zap_platforms))) {
					$serverdata['network'] = Protocol::ZOT;
				}
			}
		}

		$list = $xpath->query('//link[@rel="me"]');
		foreach ($list as $node) {
			foreach ($node->attributes as $attribute) {
				if (parse_url(trim($attribute->value), PHP_URL_HOST) == 'micro.blog') {
					$serverdata['version'] = trim($serverdata['platform'] . ' ' . $serverdata['version']);
					$serverdata['platform'] = 'microblog';
					$serverdata['network'] = Protocol::ACTIVITYPUB;
					$assigned = true;
				}
			}
		}

		if ($serverdata['platform'] != 'microblog') {
			$list = $xpath->query('//link[@rel="micropub"]');
			foreach ($list as $node) {
				foreach ($node->attributes as $attribute) {
					if (trim($attribute->value) == 'https://micro.blog/micropub') {
						$serverdata['version'] = trim($serverdata['platform'] . ' ' . $serverdata['version']);
						$serverdata['platform'] = 'microblog';
						$serverdata['network'] = Protocol::ACTIVITYPUB;
						$assigned = true;
					}
				}
			}
		}

		if ($assigned && in_array($serverdata['detection-method'], [self::DETECT_MANUAL, self::DETECT_HEADER])) {
			$serverdata['detection-method'] = self::DETECT_BODY;
		}

		return $serverdata;
	}

	/**
	 * Analyses the header data of a given server for hints about type and system of that server
	 *
	 * @param object $curlResult result of curl execution
	 * @param array  $serverdata array with server data
	 *
	 * @return array server data
	 */
	private static function analyseRootHeader($curlResult, array $serverdata): array
	{
		if ($curlResult->getHeader('server') == 'Mastodon') {
			$serverdata['platform'] = 'mastodon';
			$serverdata['network'] = Protocol::ACTIVITYPUB;
		} elseif ($curlResult->inHeader('x-diaspora-version')) {
			$serverdata['platform'] = 'diaspora';
			$serverdata['network'] = Protocol::DIASPORA;
			$serverdata['version'] = $curlResult->getHeader('x-diaspora-version')[0] ?? '';
		} elseif ($curlResult->inHeader('x-friendica-version')) {
			$serverdata['platform'] = 'friendica';
			$serverdata['network'] = Protocol::DFRN;
			$serverdata['version'] = $curlResult->getHeader('x-friendica-version')[0] ?? '';
		} else {
			return $serverdata;
		}

		if ($serverdata['detection-method'] == self::DETECT_MANUAL) {
			$serverdata['detection-method'] = self::DETECT_HEADER;
		}

		return $serverdata;
	}

	/**
	 * Update GServer entries
	 */
	public static function discover()
	{
		// Update the server list
		self::discoverFederation();

		$no_of_queries = 5;

		$requery_days = intval(DI::config()->get('system', 'poco_requery_days'));

		if ($requery_days == 0) {
			$requery_days = 7;
		}

		$last_update = date('c', time() - (60 * 60 * 24 * $requery_days));

		$gservers = DBA::select('gserver', ['id', 'url', 'nurl', 'network', 'poco', 'directory-type'],
			["NOT `failed` AND `directory-type` != ? AND `last_poco_query` < ?", GServer::DT_NONE, $last_update],
			['order' => ['RAND()']]);

		while ($gserver = DBA::fetch($gservers)) {
			Logger::info('Update peer list', ['server' => $gserver['url'], 'id' => $gserver['id']]);
			Worker::add(Worker::PRIORITY_LOW, 'UpdateServerPeers', $gserver['url']);

			Logger::info('Update directory', ['server' => $gserver['url'], 'id' => $gserver['id']]);
			Worker::add(Worker::PRIORITY_LOW, 'UpdateServerDirectory', $gserver);

			$fields = ['last_poco_query' => DateTimeFormat::utcNow()];
			self::update($fields, ['nurl' => $gserver['nurl']]);

			if (--$no_of_queries == 0) {
				break;
			}
		}

		DBA::close($gservers);
	}

	/**
	 * Discover federated servers
	 */
	private static function discoverFederation()
	{
		$last = DI::config()->get('poco', 'last_federation_discovery');

		if ($last) {
			$next = $last + (24 * 60 * 60);

			if ($next > time()) {
				return;
			}
		}

		// Discover federated servers
		$protocols = ['activitypub', 'diaspora', 'dfrn', 'ostatus'];
		foreach ($protocols as $protocol) {
			$query = '{nodes(protocol:"' . $protocol . '"){host}}';
			$curlResult = DI::httpClient()->fetch('https://the-federation.info/graphql?query=' . urlencode($query), HttpClientAccept::JSON);
			if (!empty($curlResult)) {
				$data = json_decode($curlResult, true);
				if (!empty($data['data']['nodes'])) {
					foreach ($data['data']['nodes'] as $server) {
						// Using "only_nodeinfo" since servers that are listed on that page should always have it.
						self::add('https://' . $server['host'], true);
					}
				}
			}
		}

		// Disvover Mastodon servers
		$accesstoken = DI::config()->get('system', 'instances_social_key');

		if (!empty($accesstoken)) {
			$api = 'https://instances.social/api/1.0/instances/list?count=0';
			$curlResult = DI::httpClient()->get($api, HttpClientAccept::JSON, [HttpClientOptions::HEADERS => ['Authorization' => ['Bearer ' . $accesstoken]]]);
			if ($curlResult->isSuccess()) {
				$servers = json_decode($curlResult->getBody(), true);

				if (!empty($servers['instances'])) {
					foreach ($servers['instances'] as $server) {
						$url = (is_null($server['https_score']) ? 'http' : 'https') . '://' . $server['name'];
						self::add($url);
					}
				}
			}
		}

		DI::config()->set('poco', 'last_federation_discovery', time());
	}

	/**
	 * Set the protocol for the given server
	 *
	 * @param int $gsid     Server id
	 * @param int $protocol Protocol id
	 *
	 * @throws Exception
	 */
	public static function setProtocol(int $gsid, int $protocol)
	{
		if (empty($gsid)) {
			return;
		}

		$gserver = DBA::selectFirst('gserver', ['protocol', 'url'], ['id' => $gsid]);
		if (!DBA::isResult($gserver)) {
			return;
		}

		$old = $gserver['protocol'];

		if (!is_null($old)) {
			/*
			The priority for the protocols is:
				1. ActivityPub
				2. DFRN via Diaspora
				3. Legacy DFRN
				4. Diaspora
				5. OStatus
			*/

			// We don't need to change it when nothing is to be changed
			if ($old == $protocol) {
				return;
			}

			// We don't want to mark a server as OStatus when it had been marked with any other protocol before
			if ($protocol == Post\DeliveryData::OSTATUS) {
				return;
			}

			// If the server is marked as ActivityPub then we won't change it to anything different
			if ($old == Post\DeliveryData::ACTIVITYPUB) {
				return;
			}

			// Don't change it to anything lower than DFRN if the new one wasn't ActivityPub
			if (($old == Post\DeliveryData::DFRN) && ($protocol != Post\DeliveryData::ACTIVITYPUB)) {
				return;
			}

			// Don't change it to Diaspora when it is a legacy DFRN server
			if (($old == Post\DeliveryData::LEGACY_DFRN) && ($protocol == Post\DeliveryData::DIASPORA)) {
				return;
			}
		}

		Logger::info('Protocol for server', ['protocol' => $protocol, 'old' => $old, 'id' => $gsid, 'url' => $gserver['url'], 'callstack' => System::callstack(20)]);
		self::update(['protocol' => $protocol], ['id' => $gsid]);
	}

	/**
	 * Fetch the protocol of the given server
	 *
	 * @param int $gsid Server id
	 *
	 * @return ?int One of Post\DeliveryData protocol constants or null if unknown or gserver is missing
	 *
	 * @throws Exception
	 */
	public static function getProtocol(int $gsid): ?int
	{
		if (empty($gsid)) {
			return null;
		}

		$gserver = DBA::selectFirst('gserver', ['protocol'], ['id' => $gsid]);
		if (DBA::isResult($gserver)) {
			return $gserver['protocol'];
		}

		return null;
	}

	/**
	 * Update rows in the gserver table.
	 * Enforces gserver table field maximum sizes to avoid "Data too long" database errors
	 *
	 * @param array $fields
	 * @param array $condition
	 *
	 * @return bool
	 *
	 * @throws Exception
	 */
	public static function update(array $fields, array $condition): bool
	{
		$fields = DI::dbaDefinition()->truncateFieldsForTable('gserver', $fields);

		return DBA::update('gserver', $fields, $condition);
	}

	/**
	 * Insert a row into the gserver table.
	 * Enforces gserver table field maximum sizes to avoid "Data too long" database errors
	 *
	 * @param array $fields
	 * @param int   $duplicate_mode What to do on a duplicated entry
	 *
	 * @return bool
	 *
	 * @throws Exception
	 */
	public static function insert(array $fields, int $duplicate_mode = Database::INSERT_DEFAULT): bool
	{
		$fields = DI::dbaDefinition()->truncateFieldsForTable('gserver', $fields);

		return DBA::insert('gserver', $fields, $duplicate_mode);
	}
}
